#include <EEPROM.h>
#include <flash_stm32.h>
/*
   _===_ _           _                 ______ _     _             _ _ _
  |  ___| |         | |                | ___ (_)   | |           (_) | |
  | |__ | | ___  ___| |_ _ __ ___ _____| |_/ /_  __| | ___  _   _ _| | | ___ _   _ _ __
  |  __|| |/ _ \/ __| __| '__/ _ \_____| ___ \ |/ _` |/ _ \| | | | | | |/ _ \ | | | '__|
  | |___| |  __/ (__| |_| | | (_) |    | |_/ / | (_| | (_) | |_| | | | |  __/ |_| | |
  \____/|_|\___|\___|\__|_|  \___/     \____/|_|\__,_|\___/ \__,_|_|_|_|\___|\__,_|_|


  Contrôle de l'Atténuateur HP 33322H via STM32-Mini et ULN2803. avec Affichage LCD HT1621.
  Électro-Bidouilleur, Janvier 2017
*/
//------------------------------------------------------------------------------
// Définitions et instructions au compilateur

// Broches vers l'Atténuateur
#define dB10_in_solen     PA0       // Broche d'Activation de l'atténuateur 10dB
#define dB20_in_solen     PA1       // Broche d'Activation de l'atténuateur 20dB
#define dB40_1_in_solen   PA2       // Broche d'Activation de l'atténuateur 40dB #1
#define dB40_2_in_solen   PA3       // Broche d'Activation de l'atténuateur 40dB #2
#define dB10_out_solen    PA4       // Broche de retrait de l'atténuateur 10dB
#define dB20_out_solen    PA5       // Broche de retrait de l'atténuateur 20dB
#define dB40_1_out_solen  PA6       // Broche de retrait de l'atténuateur 40dB #1
#define dB40_2_out_solen  PA7       // Broche de retrait de l'atténuateur 40dB #2
#define atten_incr_pb     PB1       // Broche du bouton poussoir d'augmentation d'Atténuation
#define atten_dec_pb      PB0       // Broche du bouton poussoir de diminution d'Atténuation
// Broches de l'affichage LCD
#define lcd_cs            PB3       // Broche CS de l"affichage LCD (fil vert)
#define lcd_wr            PB4       // Broche WR de l"affichage LCD  (fil orange)
#define lcd_data          PB5       // Broche DATA de l"affichage LCD (fil jaune)
// Autres définitions
#define pinLED PC13                 // Broche de la DEL sur la plaquette STM32 Mini
#define LF 10                       // line feed (ASCII-10)

 // Script d'alimentation d'un solénoïde
#define impuls_solen(pin)  {                            \                         
                              digitalWrite(pin, HIGH);  \
                              delay(50);                \
                              digitalWrite(pin, LOW);   \
                              delay(10);                \
                           }

 // Script de clignotement de la LED
#define flash_LED(pin) {                                \
                          digitalWrite(pin, LOW);       \
                          delay(50);                    \
                          digitalWrite(pin, HIGH);      \
                       }

 // Script de pulsation d'une broche de sortie
#define pulse_pin(pin) {                                \   
                          digitalWrite(pin, HIGH);      \
                          delayMicroseconds(10);        \
                          digitalWrite(pin, LOW);       \
                          delayMicroseconds(10);        \
                       }

//------------------------------------------------------------------------------
// Tables en mémoire flash de programme

// Table des correspondance des étages d'atténuation vs. la valeur d'atténuation en dB.
const char atten_corresp_table[] PROGMEM = {0b0000,   // 0dB
                                            0b0001,   // 10dB
                                            0b0010,   // 20dB
                                            0b0011,   // 30dB
                                            0b0100,   // 40dB
                                            0b0101,   // 50dB
                                            0b0110,   // 60dB
                                            0b0111,   // 70dB
                                            0b1100,   // 80dB
                                            0b1101,   // 90dB
                                            0b1110,   // 100dB
                                            0b1111};  // 110dB

// Table des correspondance des caractères utilisés sur l'affichage LCD Holtek HT1621 et les segments de caractères à allumer.
// Chaque caractère de l'affichage (16 positions possibles) est composé de 7 segments distribués sur 2 addresses de mémoire consécutives, une paire et une impaire
// Le bit LSB de l'addresse paire (bit "p1" ci-bas) n'est pas utilisé sur les chiffres; il sert à d'autres symboles sur l'affichage.
//
//                                      i1
//                                    -------
//                                    |     |
//                                 p2 |     | i2
//p: addresse paire                   |     |
//i: addresse impaire                 -------
//                                    | i4  |
//                                 p4 |     | i8
//                                    |     |
//                                    -------
//                                      p8
//
const char lcd_char_seg_table[] PROGMEM = { 0xE, // 0
                                            0xB,
                                            0x0, // 1
                                            0xA,
                                            0xC, // 2
                                            0x7,
                                            0x8, // 3
                                            0xF,
                                            0x2, // 4
                                            0xE,
                                            0xA, // 5
                                            0xD,
                                            0xE, // 6
                                            0xD,
                                            0x0, // 7
                                            0xB,
                                            0xE, // 8
                                            0xF,
                                            0xA, // 9
                                            0xF,
                                            0x0, // espace
                                            0x0,
                                            0xC, // d
                                            0xE,
                                            0xE, // B  (un 8)
                                            0xF };

//------------------------------------------------------------------------------
// Variables Globales

uint16 Atten_Mem_Val      = 0;              // Valeur d'atténuation courante  (à sauvegarder en mémoire Flash)
uint16 Atten_Mem_Address  = 0x10;           // Addresse de la sauvegarde d'atténuation en mémoire Flash
uint16 Status;                              // État de l'opération de sauvegarde en mémoire flash.
char Atten_dB             = 0;              // Variable intermédaire de l'Atténuation

//------------------------------------------------------------------------------
// Procédure d'activation des solénoïdes des atténuateurs

void activ_atten_solen(char nouv_atten_dB)
{
  char atten_solen_masque;                                          // Le masque utilisé pour activer/désactiver les solénoïdes

  atten_solen_masque = atten_corresp_table[nouv_atten_dB / 10];
  if (atten_solen_masque & 0b0001) impuls_solen(dB10_in_solen);     // envoi d'un bit
  if (atten_solen_masque & 0b0010) impuls_solen(dB20_in_solen);     // envoi d'un bit
  if (atten_solen_masque & 0b0100) impuls_solen(dB40_1_in_solen);   // envoi d'un bit
  if (atten_solen_masque & 0b1000) impuls_solen(dB40_2_in_solen);   // envoi d'un bit
  atten_solen_masque = ~atten_solen_masque;
  if (atten_solen_masque & 0b0001) impuls_solen(dB10_out_solen);    // envoi d'un bit
  if (atten_solen_masque & 0b0010) impuls_solen(dB20_out_solen);    // envoi d'un bit
  if (atten_solen_masque & 0b0100) impuls_solen(dB40_1_out_solen);  // envoi d'un bit
  if (atten_solen_masque & 0b1000) impuls_solen(dB40_2_out_solen);  // envoi d'un bit
}

//------------------------------------------------------------------------------
// Procédure de changement d'atténuation

void ajust_atten(int nouv_atten_dB)
{
  Serial.print("   Atténuation demandée (dB): ");                               // Afficher l'atténuation demandée sur le port série
  Serial.println(nouv_atten_dB);
  Atten_Mem_Val = nouv_atten_dB;                                                // Sauvegarder la nouvelle atténuation en mémoire flash
  Status = EEPROM.write(Atten_Mem_Address, Atten_Mem_Val);                      //      "
  if (Status == 0) Serial.println("   Atténuation sauvegardée en mémoire.");    // Si sauvegarde réussie, afficher confirmation de sauvegarde sur le port série
  else Serial.println("   Échec de sauvegarde en mémoire de l'atténuation.");   // Sinon, afficher échec de sauvegarde sur le port série
  activ_atten_solen(nouv_atten_dB);                                             // Activer les solénoïdes selon l'atténuation demandée.
  affich_atten(Atten_Mem_Val);                                                  // Mettre à jour de l'affichage LCD avec la nouvelle atténuation
}
// -------------------------------------------------------------------------------
// Procédure d'envoi de segments de caractère à allumer à une addresse de l'affichage LCD

void allum_seg(char address,char commum)
{
  int pos,k;                                      // Variables temporaires
  digitalWrite(lcd_cs, LOW);                      // Abaisser le CS
  delayMicroseconds(10);
  digitalWrite(lcd_data, HIGH);                   // Envoyer 101 (préambule d'une commande d'affichage)
  delayMicroseconds(10);
  pulse_pin(lcd_wr);
  digitalWrite(lcd_data, LOW);                    //      "
  delayMicroseconds(10);
  pulse_pin(lcd_wr);
  digitalWrite(lcd_data, HIGH);                   //      "
  delayMicroseconds(10);
  pulse_pin(lcd_wr);
  for (pos = 5; pos >= 0; pos--)                  // Boucle d'envoi de l'adresse de la mémoire de caractère (0-31) à modifier, MSB en premier
  {
      k = address >> pos;
     if (k & 1) digitalWrite(lcd_data, HIGH);
      else digitalWrite(lcd_data, LOW);
      delayMicroseconds(10);
      pulse_pin(lcd_wr);
  }
   // caractère
  for (pos = 3; pos >= 0; pos--)                  // Boucle d'envoi des segments de caractères (lignes communes 0-3) en mémoire du LCD, MSB en premier
  {
      k = commum >> pos;
     if (k & 1) digitalWrite(lcd_data, HIGH);
      else digitalWrite(lcd_data, LOW);
      delayMicroseconds(10);
      pulse_pin(lcd_wr);
  }
  digitalWrite(lcd_cs, HIGH);                     // Relever le CS
  delayMicroseconds(10);
}

// -------------------------------------------------------------------------------
// Procédure d'envoi d'une commande de contrôle à l'affichage LCD

void env_commande(uint16 data)
{
  int pos,k;
  digitalWrite(lcd_cs, LOW);                      // Abaisser le CS
  delayMicroseconds(10);
  digitalWrite(lcd_data, HIGH);                   // Envoyer 100 (préambule d'une commande de contrôle)
  delayMicroseconds(10);
  pulse_pin(lcd_wr);
  digitalWrite(lcd_data, LOW);                    //      "
  delayMicroseconds(10);
  pulse_pin(lcd_wr);
  digitalWrite(lcd_data, LOW);                    //      "
  delayMicroseconds(10);
  pulse_pin(lcd_wr);
  for (pos = 8; pos >= 0; pos--)                  // Boucle d'envoi des 9 bits de contrôle, un bit à la fois, MSB en premier
  {
      k = data >> pos;
     if (k & 1) digitalWrite(lcd_data, HIGH);
      else digitalWrite(lcd_data, LOW);
      delayMicroseconds(10);
      pulse_pin(lcd_wr);
  }
  digitalWrite(lcd_cs, HIGH);                     // Relever le CS
  delayMicroseconds(10);
}

// -------------------------------------------------------------------------------
// Procédure de mise à jour de l'atténuation sur l'affichage LCD

void affich_atten(char atten)
{
  char i;                                             // Variable temporaire
  
  for (i = 0; i <= 6; i=i+2)                          // Ajouter des espaces pour débuter
  {
    allum_seg(i, lcd_char_seg_table[20]);             // espace
    allum_seg(i+1, lcd_char_seg_table[21]);
  }
  i++;
  if (atten >= 100)                                   // Traiter le cas où l'atténuation est 100dB ou plus
  {
    allum_seg(++i, lcd_char_seg_table[2]);            // Ajouter un "1"
    allum_seg(++i, lcd_char_seg_table[3]);
    atten = atten - 100;                              // Soustraire 100dB 
  }
  else                                                // Traiter le cas où l'atténuation est moins de 100dB
  {
    allum_seg(++i, lcd_char_seg_table[20]);           // Ajouter un espace
    allum_seg(++i, lcd_char_seg_table[21]);
  }
  allum_seg(++i, lcd_char_seg_table[atten/10*2]);     // Ajouter le chiffre des dizaines de dB
  allum_seg(++i, lcd_char_seg_table[(atten/10*2)+1]);
  allum_seg(++i, lcd_char_seg_table[0]);              // Ajouter un "0"
  allum_seg(++i, lcd_char_seg_table[1]);
  allum_seg(++i, lcd_char_seg_table[20]);             // Ajouter un espace
  allum_seg(++i, lcd_char_seg_table[21]);
  allum_seg(++i, lcd_char_seg_table[22]);             // Ajouter un "d"
  allum_seg(++i, lcd_char_seg_table[23]);   
  allum_seg(++i, lcd_char_seg_table[16]);             // Ajouter un "8" pour émuler un "B"
  allum_seg(++i, lcd_char_seg_table[17]);
}

//------------------------------------------------------------------------------
// Fonction de vérification de nombre. Retourne Vrai si la chaîne de caractères représente un nombre

boolean IsNumeric(String str)
{
  if (str.length() == 0)  return false;               // si la chaîne est vide, pas un nombre
  for (char i = 0; i < str.length(); i++)             // tester tous les caractères, un à la fois
  {
    if (!(isDigit(str.charAt(i)) ))                   // Est-ce que le caractère est un chiffre?
    {
      return false;                                   // Non, retourner Faux
    }
  }
  return true;                                        // Ici, tous les caractères ont été vérifiés comme étant des chiffres. Retourner Vrai.
}

//------------------------------------------------------------------------------
// Procédure de configuration du STM32, style arduino

void setup()
{
  char i;                                             // Variable-compteur
  
   // configurations des broches divers
  pinMode(dB10_in_solen, OUTPUT);                     // Sortie
  pinMode(dB20_in_solen, OUTPUT);                     // Sortie
  pinMode(dB40_1_in_solen, OUTPUT);                   // Sortie
  pinMode(dB40_2_in_solen, OUTPUT);                   // Sortie
  pinMode(dB10_out_solen, OUTPUT);                    // Sortie
  pinMode(dB20_out_solen, OUTPUT);                    // Sortie
  pinMode(dB40_1_out_solen, OUTPUT);                  // Sortie
  pinMode(dB40_2_out_solen, OUTPUT);                  // Sortie
  pinMode(atten_incr_pb, INPUT);                      // Entrée
  pinMode(atten_dec_pb, INPUT);                       // Entrée
  pinMode(pinLED, OUTPUT);                            // Sortie

  digitalWrite(pinLED, HIGH);                         // éteindre la DEL

  Serial.begin(9600);                                 // configuration du port série (broches A9 et A10)
  Serial.setTimeout(10);                              // Délai d'attente maximum de la fonction serial.read

   // Définir la zone du flash à utiliser comme émulation de EEPROM
  EEPROM.PageBase0 = 0x801F000;
  EEPROM.PageBase1 = 0x801F800;
  EEPROM.PageSize  = 0x400;

   // Mettre à jour l'atténuation selon la valeur sauvegardée.
  Status = EEPROM.read(Atten_Mem_Address, &Atten_Mem_Val);                                  // Lire la valeur sauvegardée en mémoire flash
  if ((Atten_Mem_Val < 0) || (Atten_Mem_Val > 110) || (Status != 0)) Atten_Mem_Val = 110;   // couvre les cas où la mémoire serait vide ou corrompue.
  Serial.print("   Atténuation en Mémoire (dB): ");                                         // Afficher l'atténuation sur le port série
  Serial.println(Atten_Mem_Val);                                                            //          "
  activ_atten_solen(Atten_Mem_Val);                                                         // Activer les solénoïdes selon l'atténuation en mémoire. S'assure que la valeur d'atténuation sauvegardée est bien celle appliquée
  Serial.println("Entrez l'atténuation (dB): ");                                            // Afficher la demande d'atténuation à l'usager sur le port série

   // Initialisation de l'Affichage
  pinMode(lcd_cs, OUTPUT);                                // Chip Select
  pinMode(lcd_wr, OUTPUT);                                // Write
  pinMode(lcd_data, OUTPUT);                              // Data
  digitalWrite(lcd_wr, LOW);                              // Mettre la broche WR au niveau bas
  digitalWrite(lcd_cs, HIGH);                             // Soulever le CS
  delay(10);                                              // Pause de 10 millisecondes
   // Configuration de l'affichage
  env_commande(0b000000010);                              // Envoyer la commande d'allumage de l'oscillateur de système du LCD
  env_commande(0b000000111);                              // Envoyer la commande d'allumage du générateur de Bias du LCD
  env_commande(0b001010111);                              // Envoyer la commande d'ajustement du Bias et dunombre de lignes communes
   
  for (i = 0; i <= 31; i++) allum_seg(i, 0x0);            // Effacer toutes les cellules de memoire (tous les caractères)
  affich_atten(Atten_Mem_Val);                            // Afficher l'atténuation initiale.

//   // DURANT LE DÉVELOPPEMENT, Fait allumer les segments individuellement pour pouvoir les identifier
//  allum_seg(0, 1);        // 0
//  allum_seg(2, 2);
//  allum_seg(4, 4);        // 1
//  allum_seg(6, 8);
//  allum_seg(9, 1);        // 2
//  allum_seg(11, 2);
//  allum_seg(13, 4);       // 3
//  allum_seg(15, 8);
//  while(1);
}

//------------------------------------------------------------------------------
// Boucle principale, style arduino

void loop()
{
   // Définition des variables
  String atten_input = "";                                            // La chaîne de caractères contenant la fréquence
  char temp_char;                                                     // Caractère temporaire

   // Boucle principale
  while (true)                                                        // Boucle perpétuelle
  {
     // Validation d'un changement d'atténuation par port série
    temp_char = Serial.read();                                        // Lire un caractère du port série.  Expiration de la lecture après 10ms
    if (temp_char >= 10 && temp_char <= 127)                          // Validation du caractère lu (est-ce un caractère?)
    {
      if (temp_char != LF) atten_input += temp_char;                  // Ajouter le caractère à la chaîne de commande si la fin de la commande (délimitée par LF (ASCII 10)) n'est pas reçue
      else                                                            // Fin de la commande reçue ici
      {
        Serial.print("   Atténuation entrée (dB): ");                 // Afficher sur le port série l'atténuation reçue 
        Serial.println(atten_input);                                  //       "
        if ((IsNumeric(atten_input)) && (atten_input.toInt() >= 0) && (atten_input.toInt() <= 110) && (atten_input.toInt() % 10 == 0))    // Valider si la chaîne de caractères représente un nombre de 0 à 110 en dizaine?
        {
          ajust_atten(atten_input.toInt());                           // Atténuation valide. Ajuster l'atténuation à la nouvelle valeur
          flash_LED(pinLED);                                          // Faire clignoter le LED
        }
        else Serial.println("   Erreur dans le format de l'atténuation entrée. Veuillez recommencer...");   // Sinon, mauvaise valeur d'atténuation reçue
        atten_input = "";                                             // Ré-initialiser la chaîne de caractères à vide.
        Serial.println("Entrez l'atténuation (dB): ");                // Afficher la demande d'atténuation à l'usager sur le port série
      }
    }
     // Validation d'un changement d'atténuation par bouton poussoir
    if (!digitalRead(atten_incr_pb) || !digitalRead(atten_dec_pb))    // Un des bouton enfoncé ?
    {
      flash_LED(pinLED);                                              // Oui, faire clignoter le DEL
      if (!digitalRead(atten_incr_pb) && (Atten_Mem_Val < 110))       // Bouton d'augmentation enfoncé?
      { 
        ajust_atten(Atten_Mem_Val + 10);                              // Oui, augmenter l'atténuation
      }
      else if (!digitalRead(atten_dec_pb) && Atten_Mem_Val > 0)       // Non, Bouton de diminution enfoncé?
      {
        ajust_atten(Atten_Mem_Val - 10);                              // Oui, diminuer l'atténuation
      }
      Serial.println("Entrez l'atténuation (dB): ");                  // Afficher la demande d'atténuation à l'usager sur le port série
    }
  }
}
